@@ -3,7 +3,7 @@
 # This file is part of Radicale Server - Calendar Server
 # Copyright © 2008 Nicolas Kandel
 # Copyright © 2008 Pascal Halter
-# Copyright © 2008-2013 Guillaume Ayoub
+# Copyright © 2008-2015 Guillaume Ayoub
 #
 # This library is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -32,7 +32,6 @@ import os
 import sys
 import pprint
 import base64
-import posixpath
 import socket
 import ssl
 import wsgiref.simple_server
@@ -48,16 +47,16 @@ except ImportError:
     from urlparse import urlparse
 # pylint: enable=F0401,E0611
 
-from . import auth, config, ical, log, rights, storage, xmlutils
+from . import auth, config, ical, log, pathutils, rights, storage, xmlutils
 
 
-VERSION = "0.10"
+VERSION = "1.1"
 
 # Standard "not allowed" response that is returned when an authenticated user
 # tries to access information they don't have rights to
 NOT_ALLOWED = (client.FORBIDDEN, {}, None)
 
-WELL_KNOWN_RE = re.compile(r"/.well-known/(carddav|caldav)/?")
+WELL_KNOWN_RE = re.compile(r"/\.well-known/(carddav|caldav)/?$")
 
 
 class HTTPServer(wsgiref.simple_server.WSGIServer, object):
@@ -159,7 +158,8 @@ class Application(object):
         # First append content charset given in the request
         content_type = environ.get("CONTENT_TYPE")
         if content_type and "charset=" in content_type:
-            charsets.append(content_type.split("charset=")[1].strip())
+            charsets.append(
+                content_type.split("charset=")[1].split(";")[0].strip())
         # Then append default Radicale charset
         charsets.append(self.encoding)
         # Then append various fallbacks
@@ -176,12 +176,9 @@ class Application(object):
 
     @staticmethod
     def sanitize_uri(uri):
-        """Unquote and remove /../ to prevent access to other data."""
+        """Unquote and make absolute to prevent access to other data."""
         uri = unquote(uri)
-        trailing_slash = "/" if uri.endswith("/") else ""
-        uri = posixpath.normpath(uri)
-        trailing_slash = "" if uri == "/" else trailing_slash
-        return uri + trailing_slash
+        return pathutils.sanitize_path(uri)
 
     def collect_allowed_items(self, items, user):
         """Get items from request that user is allowed to access."""
@@ -248,25 +245,29 @@ class Application(object):
         headers = pprint.pformat(self.headers_log(environ))
         log.LOGGER.debug("Request headers:\n%s" % headers)
 
+        # Strip base_prefix from request URI
         base_prefix = config.get("server", "base_prefix")
         if environ["PATH_INFO"].startswith(base_prefix):
-            # Sanitize request URI
-            environ["PATH_INFO"] = self.sanitize_uri(
-                "/%s" % environ["PATH_INFO"][len(base_prefix):])
-            log.LOGGER.debug("Sanitized path: %s", environ["PATH_INFO"])
+            environ["PATH_INFO"] = environ["PATH_INFO"][len(base_prefix):]
         elif config.get("server", "can_skip_base_prefix"):
             log.LOGGER.debug(
-                "Skipped already sanitized path: %s", environ["PATH_INFO"])
+                "Prefix already stripped from path: %s", environ["PATH_INFO"])
         else:
             # Request path not starting with base_prefix, not allowed
             log.LOGGER.debug(
                 "Path not starting with prefix: %s", environ["PATH_INFO"])
-            environ["PATH_INFO"] = None
+            status, headers, _ = NOT_ALLOWED
+            start_response(status, list(headers.items()))
+            return []
+
+        # Sanitize request URI
+        environ["PATH_INFO"] = self.sanitize_uri(environ["PATH_INFO"])
+        log.LOGGER.debug("Sanitized path: %s", environ["PATH_INFO"])
 
         path = environ["PATH_INFO"]
 
         # Get function corresponding to method
-        function = getattr(self, environ["REQUEST_METHOD"].lower())
+        function = getattr(self, "do_%s" % environ["REQUEST_METHOD"].upper())
 
         # Ask authentication backend to check rights
         authorization = environ.get("HTTP_AUTHORIZATION", None)
@@ -324,8 +325,8 @@ class Application(object):
 
         if is_valid_user and (
                 (read_allowed_items or write_allowed_items) or
-                (is_authenticated and function == self.propfind) or
-                function == self.options):
+                (is_authenticated and function == self.do_PROPFIND) or
+                function == self.do_OPTIONS):
             status, headers, answer = function(
                 environ, read_allowed_items, write_allowed_items, content,
                 user)
@@ -364,8 +365,8 @@ class Application(object):
     # All these functions must have the same parameters, some are useless
     # pylint: disable=W0612,W0613,R0201
 
-    def delete(self, environ, read_collections, write_collections, content,
-               user):
+    def do_DELETE(self, environ, read_collections, write_collections, content,
+                  user):
         """Manage DELETE request."""
         if not len(write_collections):
             return NOT_ALLOWED
@@ -377,13 +378,13 @@ class Application(object):
             item = collection
         else:
             # Try to get an item matching the path
-            item = collection.get_item(
-                xmlutils.name_from_path(environ["PATH_INFO"], collection))
+            name = xmlutils.name_from_path(environ["PATH_INFO"], collection)
+            item = collection.items.get(name)
 
         if item:
             # Evolution bug workaround
-            etag = environ.get("HTTP_IF_MATCH", item.etag).replace("\\", "")
-            if etag == item.etag:
+            if_match = environ.get("HTTP_IF_MATCH", "*").replace("\\", "")
+            if if_match in ("*", item.etag):
                 # No ETag precondition or precondition verified, delete item
                 answer = xmlutils.delete(environ["PATH_INFO"], collection)
                 return client.OK, {}, answer
@@ -391,7 +392,8 @@ class Application(object):
         # No item or ETag precondition not verified, do not delete item
         return client.PRECONDITION_FAILED, {}, None
 
-    def get(self, environ, read_collections, write_collections, content, user):
+    def do_GET(self, environ, read_collections, write_collections, content,
+               user):
         """Manage GET request.
 
         In Radicale, GET requests create collections when the URL is not
@@ -414,15 +416,16 @@ class Application(object):
 
         if item_name:
             # Get collection item
-            item = collection.get_item(item_name)
+            item = collection.items.get(item_name)
             if item:
-                items = collection.timezones
-                items.append(item)
+                items = [item]
+                if collection.resource_type == "calendar":
+                    items.extend(collection.timezones)
                 answer_text = ical.serialize(
                     collection.tag, collection.headers, items)
                 etag = item.etag
             else:
-                return client.GONE, {}, None
+                return client.NOT_FOUND, {}, None
         else:
             # Create the collection if it does not exist
             if not collection.exists:
@@ -447,15 +450,15 @@ class Application(object):
         answer = answer_text.encode(self.encoding)
         return client.OK, headers, answer
 
-    def head(self, environ, read_collections, write_collections, content,
-             user):
+    def do_HEAD(self, environ, read_collections, write_collections, content,
+                user):
         """Manage HEAD request."""
         status, headers, answer = self.get(
             environ, read_collections, write_collections, content, user)
         return status, headers, None
 
-    def mkcalendar(self, environ, read_collections, write_collections, content,
-                   user):
+    def do_MKCALENDAR(self, environ, read_collections, write_collections,
+                      content, user):
         """Manage MKCALENDAR request."""
         if not len(write_collections):
             return NOT_ALLOWED
@@ -473,8 +476,8 @@ class Application(object):
             collection.write()
         return client.CREATED, {}, None
 
-    def mkcol(self, environ, read_collections, write_collections, content,
-              user):
+    def do_MKCOL(self, environ, read_collections, write_collections, content,
+                 user):
         """Manage MKCOL request."""
         if not len(write_collections):
             return NOT_ALLOWED
@@ -488,8 +491,8 @@ class Application(object):
         collection.write()
         return client.CREATED, {}, None
 
-    def move(self, environ, read_collections, write_collections, content,
-             user):
+    def do_MOVE(self, environ, read_collections, write_collections, content,
+                user):
         """Manage MOVE request."""
         if not len(write_collections):
             return NOT_ALLOWED
@@ -499,7 +502,7 @@ class Application(object):
         from_name = xmlutils.name_from_path(
             environ["PATH_INFO"], from_collection)
         if from_name:
-            item = from_collection.get_item(from_name)
+            item = from_collection.items.get(from_name)
             if item:
                 # Move the item
                 to_url_parts = urlparse(environ["HTTP_DESTINATION"])
@@ -524,8 +527,8 @@ class Application(object):
             # Moving collections, not supported
             return client.FORBIDDEN, {}, None
 
-    def options(self, environ, read_collections, write_collections, content,
-                user):
+    def do_OPTIONS(self, environ, read_collections, write_collections,
+                   content, user):
         """Manage OPTIONS request."""
         headers = {
             "Allow": ("DELETE, HEAD, GET, MKCALENDAR, MKCOL, MOVE, "
@@ -533,8 +536,8 @@ class Application(object):
             "DAV": "1, 2, 3, calendar-access, addressbook, extended-mkcol"}
         return client.OK, headers, None
 
-    def propfind(self, environ, read_collections, write_collections, content,
-                 user):
+    def do_PROPFIND(self, environ, read_collections, write_collections,
+                    content, user):
         """Manage PROPFIND request."""
         # Rights is handled by collection in xmlutils.propfind
         headers = {
@@ -545,8 +548,8 @@ class Application(object):
             environ["PATH_INFO"], content, collections, user)
         return client.MULTI_STATUS, headers, answer
 
-    def proppatch(self, environ, read_collections, write_collections, content,
-                  user):
+    def do_PROPPATCH(self, environ, read_collections, write_collections,
+                     content, user):
         """Manage PROPPATCH request."""
         if not len(write_collections):
             return NOT_ALLOWED
@@ -560,7 +563,8 @@ class Application(object):
             "Content-Type": "text/xml"}
         return client.MULTI_STATUS, headers, answer
 
-    def put(self, environ, read_collections, write_collections, content, user):
+    def do_PUT(self, environ, read_collections, write_collections, content,
+               user):
         """Manage PUT request."""
         if not len(write_collections):
             return NOT_ALLOWED
@@ -570,7 +574,7 @@ class Application(object):
         collection.set_mimetype(environ.get("CONTENT_TYPE"))
         headers = {}
         item_name = xmlutils.name_from_path(environ["PATH_INFO"], collection)
-        item = collection.get_item(item_name)
+        item = collection.items.get(item_name)
 
         # Evolution bug workaround
         etag = environ.get("HTTP_IF_MATCH", "").replace("\\", "")
@@ -587,7 +591,7 @@ class Application(object):
             # If the added item doesn't have the same name as the one given
             # by the client, then there's no obvious way to generate an
             # etag, we can safely ignore it.
-            new_item = collection.get_item(item_name)
+            new_item = collection.items.get(item_name)
             if new_item:
                 headers["ETag"] = new_item.etag
         else:
@@ -595,8 +599,8 @@ class Application(object):
             status = client.PRECONDITION_FAILED
         return status, headers, None
 
-    def report(self, environ, read_collections, write_collections, content,
-               user):
+    def do_REPORT(self, environ, read_collections, write_collections, content,
+                  user):
         """Manage REPORT request."""
         if not len(read_collections):
             return NOT_ALLOWED
